/***********************************************************************************
 * LabTestSim - Laboratory Test Simulator using Elastoplastic models               *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>     *
 *                                                                                 *
 * This file is part of LabTestSim.                                                *
 *                                                                                 *
 * LabTestSim is free software; you can redistribute it and/or modify it under the *
 * terms of the GNU General Public License as published by the Free Software       *
 * Foundation; either version 2 of the License, or (at your option) any later      *
 * version.                                                                        *
 *                                                                                 *
 * LabTestSim is distributed in the hope that it will be useful, but WITHOUT ANY   *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.        *
 *                                                                                 *
 * You should have received a copy of the GNU General Public License along with    *
 * LabTestSim; if not, write to the Free Software Foundation, Inc., 51 Franklin    *
 * Street, Fifth Floor, Boston, MA 02110-1301, USA                                 *
 ***********************************************************************************/

#include <iostream>
#include <iomanip>
#include <fstream>
#include "labtestsim.h"

// external
#include "mechsys/parameters/parameters.h"
#include "mechsys/models/model.h"
#include "mechsys/base/mechsys.h"
#include "mechsys/invariants/stressinvs.h"
#include "mechsys/invariants/straininvs.h"

bool LabTestSim::Set(const char          * strTestKey,
					 MechSys::Parameters * Prms,
	                 MechSys::Model      * Model,
	                 std::string           PatFileName,
	                 int                   NumStep,
	                 int                   OutStep,
				     double                MaxEdPct)
{
	_key = _char_to_key(strTestKey);
	if (_key==NON)
	{
		_str_errors.append("LabTestSim::Set::ERROR: Invalid TestKey !!!\n");
		return false;
	}
	   _model = Model;
	_num_step = NumStep;
	_out_step = OutStep;
	  _max_ed = MaxEdPct/100.0;

	// Initial Strain state
	MechSys::Tensr2 Eps_0;
	_eps.push_back(Eps_0);

	// Path
	std::ifstream file(PatFileName.c_str(),std::ios::in);
	if (file)
	{
		// Heading
		int nlines_head = 2;
		std::string head[nlines_head];
		for (int i=0; i<nlines_head; ++i)
			std::getline(file,head[i]);

		double Sa;
		double Sr;
		double XX;

		// Initial Stress
		MechSys::Tensr2 Sig_0;
		file>>Sa>>Sr>>XX;
		Sig_0[0] = Sa;
		Sig_0[1] = Sr;
		Sig_0[2] = Sr;
		_sig.push_back(Sig_0);
		_sgnX.push_back(+1.0);
		_sgnY.push_back(+1.0);

		// Parameters
		Prms->Calculatevri(Sig_0);
		
		// Model
		if (!_model->Set(Prms,Sig_0)) return false;

		// Data
		_num_path = 0;
		while (file>>Sa>>Sr>>XX)
		{
			_path_Sa.push_back(Sa);
			_path_Sr.push_back(Sr);
			_path_XX.push_back(XX);
			_num_path++;
		}
		file.close();
		return true;
	}
	else
	{
		_str_errors.append("LabTestSim::Set::ERROR: Could not open file <");
		_str_errors.append(PatFileName);
		_str_errors.append("> !!!\n");
		return false;
	}
}


bool LabTestSim::RunTest()
{
	MechSys::Tensr2 Sig(_sig.back());
	MechSys::Tensr2 Eps;
	MechSys::Tensr2 DSig;
	MechSys::Tensr2 DEps;

	switch (_key)
	{
	case NON:
		{
			_str_errors.append("LabTestSim::RunTest::ERROR: TestKey not specified !!!\n");
			return false;
		}
	case CTX:
		break;
	case CTP: // XX=Ea
		{
			double DEa;  // Vertical (a) strain increment
			for (int i=0; i<_num_path; ++i)
			{
				DSig.Clear();
				DEps.Clear();
				DSig[0] = (_path_Sa[i] - Sig[0])/_num_step;
				DSig[1] = (_path_Sr[i] - Sig[1])/_num_step;
				DSig[2] = (_path_Sr[i] - Sig[2])/_num_step;
				DEa     = _path_XX[i]/_num_step;
				int step = 0;
				for (int j=0; j<_num_step; ++j)
				{
                    if (_path_XX[i]==0)              // Stress controlled
					{
                        if (!_model->Actualize(DSig,DEps))
						{
							_str_errors.append("LabTestSim::RunTest::ERROR: Could not Actualize model !!!\n");
							return false;
						}
						Sig = Sig + DSig;
						Eps = Eps + DEps;
					}
                    else                             // Strain controlled
					{
						MechSys::Tensr4 * C = _model->TgCompliance();
                        DSig[0] = DEa/((*C)[0][0]-0.5*((*C)[0][1]+(*C)[0][2]));         // DSa
                        DSig[1] = -0.5*DSig[0];                                         // DSr
                        DSig[2] = DSig[1];                                              // DSt
                        DEps[0] = DEa;                                                  // DEa
                        DEps[1] = (*C)[1][0]*DSig[0] + ((*C)[1][1]+(*C)[1][2])*DSig[1]; // DEr
                        DEps[2] = DEps[1];                                              // DEt
						Sig = Sig + DSig;
						Eps = Eps + DEps;
						if (!_model->Actualize(DSig,DEps))
						{
							_str_errors.append("LabTestSim::RunTest::ERROR: Could not Actualize model !!!\n");
							return false;
						}
					}
					if (step==_out_step-1 || j==_num_step-1)
					{
						_sig.push_back(Sig);
						_eps.push_back(Eps);
						if (Eps[0]>=Eps[1]) _sgnX.push_back(+1.0);
						else                _sgnX.push_back(-1.0);
						if (Sig[0]>=Sig[1]) _sgnY.push_back(+1.0);
						else                _sgnY.push_back(-1.0);
						step = -1;
					}
					step++;
					MechSys::StrainInvs ei(Eps);
					double ev;
					double ed;
					ei.Get_ev_ed(ev,ed);
					if (i==_num_path-1 & fabs(ed)>_max_ed) break;
				}
			}
			break;
		}
	case CTA:
		break;
	case CTR:
		break;
	case TTX:
		break;
	case TTP:
		break;
	case HCY:
		break;
	case PSA:
		break;
	}

	return true;

} // RunTest()


TestKey LabTestSim::_char_to_key(const char * strTestKey)
{
	if (strcmp(strTestKey,"CTX")==0) return CTX;
	if (strcmp(strTestKey,"CTP")==0) return CTP;
	if (strcmp(strTestKey,"CTA")==0) return CTA;
	if (strcmp(strTestKey,"CTR")==0) return CTR;
	if (strcmp(strTestKey,"TTX")==0) return TTX;
	if (strcmp(strTestKey,"TTP")==0) return TTP;
	if (strcmp(strTestKey,"HCY")==0) return HCY;
	if (strcmp(strTestKey,"PSA")==0) return PSA;
	return NON;
}


std::ostream & operator<< (std::ostream & os, const LabTestSim & lts)
{
	// Caption
	os << std::setw(12) << "Sa";
	os << std::setw(12) << "Sr";
	os << std::setw(12) << "St";
	os << std::setw(12) << "Sat";
	os << std::setw(12) << "Ea(%)";
	os << std::setw(12) << "Er(%)";
	os << std::setw(12) << "Et(%)";
	os << std::setw(12) << "Eat(%)";
	os << std::setw(12) << "p";
	os << std::setw(12) << "q";
	os << std::setw(12) << "ev(%)";
	os << std::setw(12) << "ed(%)";
	os << std::setw(12) << "sgnX";
	os << std::setw(12) << "sgnY";
	os << std::endl;

	MechSys::StressInvs si(lts._sig[0]);
	MechSys::StrainInvs ei(lts._eps[0]);

	double p;
	double q;
	double ev;
	double ed;

	for (int i=0; i<static_cast<int>(lts._sig.size()); ++i)
	{

		// Stress
		for (int j=0; j<4; ++j)
			os << std::setw(12) << std::showpos << std::fixed << std::setprecision(6) << lts._sig[i][j];

		// Strain
		for (int j=0; j<4; ++j)
			os << std::setw(12) << std::showpos << std::fixed << std::setprecision(6) << lts._eps[i][j]*100.0;

		// Invariants
		si.Reset(lts._sig[i]);
		ei.Reset(lts._eps[i]);
		si.Get_p_q(p,q);
		ei.Get_ev_ed(ev,ed);
		os << std::setw(12) << std::showpos << std::fixed << std::setprecision(6) << p;
		os << std::setw(12) << std::showpos << std::fixed << std::setprecision(6) << q;
		os << std::setw(12) << std::showpos << std::fixed << std::setprecision(6) << ev*100.0;
		os << std::setw(12) << std::showpos << std::fixed << std::setprecision(6) << ed*100.0;

		// Signal to help plot
		os << std::setw(12) << std::fixed << std::setprecision(6) << lts._sgnX[i];
		os << std::setw(12) << std::fixed << std::setprecision(6) << lts._sgnY[i];
		
		// Newline
		os << std::endl;
	}
	
	return os;
} // operator <<
