/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

#include <cstring>
#include <cassert>
#include <cmath>
#include "mechsys.h"

//external
#include "../linalg/matrix.h"
#include "../linalg/jacobirot.h"

// Helpers
MechSys::ModelID MechSys::GetModelID(const char * model)
{
	if (strcmp(model,"subcam")==0) return MechSys::subcam;
	if (strcmp(model,"subtij")==0) return MechSys::subtij;
	return MechSys::wrongmodel;
}

MechSys::SchemeID MechSys::GetSchemeID(const char * scheme)
{
	if (strcmp(scheme,"FE")==0) return MechSys::FE;
	if (strcmp(scheme,"ME")==0) return MechSys::ME;
	return MechSys::wrongscheme;
}

// Functions
void MechSys::tensor_to_matrix(const MechSys::Tensr2 & T,
		                             LinAlg::Matrix  & M)
{
	double sq2 = sqrt(2.0);
	for (int i=0; i<3; ++i) M(i,i) = T[i];
	M(0,1) = T[3]/sq2; M(1,0) = M(0,1);
	M(1,2) = T[4]/sq2; M(2,1) = M(1,2);
	M(0,2) = T[5]/sq2; M(2,0) = M(0,2);
}

void MechSys::matrix_to_tensor(const LinAlg::Matrix  & M,
		                             MechSys::Tensr2 & T)
{
	double sq2 = sqrt(2.0);
	for (int i=0; i<3; ++i) T[i] = M(i,i);
	T[3] = M(0,1)*sq2;
	T[4] = M(1,2)*sq2;
	T[5] = M(0,2)*sq2;
}

// Tensor multiplication:  ST = S*T
void MechSys::tensor_mult(const MechSys::Tensr2 & S,
		                  const MechSys::Tensr2 & T,
						        MechSys::Tensr2 & ST)
{
	// -*- IMPORTANT -*-
	//       The function tensor_mult is valid only if the result R=S*T is 
	//       symmetric. This is NOT always true, even if S and T are both
	//       symmetric.

	double sq2 = sqrt(2.0);

	ST[0] =  S[0]*T[0]     + S[3]*T[3]/2.0 + S[5]*T[5]/2.0;
	ST[1] =  S[3]*T[3]/2.0 + S[1]*T[1]     + S[4]*T[4]/2.0;
	ST[2] =  S[5]*T[5]/2.0 + S[4]*T[4]/2.0 + S[2]*T[2];

	ST[3] = S[0]*T[3]     + S[3]*T[1]     + S[5]*T[4]/sq2;
	ST[4] = S[3]*T[5]/sq2 + S[1]*T[4]     + S[4]*T[2];
	ST[5] = S[0]*T[5]     + S[3]*T[4]/sq2 + S[5]*T[2];

#ifndef NDEBUG
	const double ZERO = 1.0e-10; // to consider a tensor symmetric
	double A = S[3]*T[0]     + S[1]*T[3]     + S[4]*T[5]/sq2;
	double B = S[5]*T[3]/sq2 + S[4]*T[1]     + S[2]*T[4];
	double C = S[5]*T[0]     + S[4]*T[3]/sq2 + S[2]*T[5];
	assert(fabs(ST[3]-A)<ZERO & fabs(ST[4]-B)<ZERO & fabs(ST[5]-C)<ZERO);
#endif
}

// Inverse of a tensor T:  invT = T^(-1)
bool MechSys::tensor_inv(const MechSys::Tensr2 & T,
		                       MechSys::Tensr2 & invT)
{
	double ZERO = 1.0e-14;
	double  sq2 = sqrt(2.0);

	LinAlg::Matrix M(3,3);
	MechSys::tensor_to_matrix(T,M);
	
	double det =  M(0,0)*M(1,1)*M(2,2)
				+ M(0,1)*M(1,2)*M(2,0)
				+ M(0,2)*M(2,1)*M(1,0) 
				- M(0,0)*M(1,2)*M(2,1)
				- M(1,1)*M(0,2)*M(2,0)
				- M(2,2)*M(0,1)*M(1,0);
	
	if (fabs(det)<ZERO) return false;

	invT[0] =     (M(1,1)*M(2,2)-M(1,2)*M(2,1))/det;
	invT[1] =     (M(2,2)*M(0,0)-M(2,0)*M(0,2))/det;
	invT[2] =     (M(0,0)*M(1,1)-M(0,1)*M(1,0))/det;
	invT[3] = sq2*(M(2,1)*M(0,2)-M(2,2)*M(0,1))/det;
	invT[4] = sq2*(M(0,2)*M(1,0)-M(0,0)*M(1,2))/det;
	invT[5] = sq2*(M(0,1)*M(1,2)-M(0,2)*M(1,1))/det;

	return true;
}

// Eigenvalues (ev) and Eigenprojectors (ep) of a tensor T
bool MechSys::tensor_eigenvp(const MechSys::Tensr2 & T,
		                           double            ev[3],
								   MechSys::Tensr2   ep[3])
{
	double sq2 = sqrt(2.0);
	LinAlg::Matrix M(3,3); // Matrix of tensor T
	LinAlg::Matrix Q(3,3); // Orthogonal matrix for tensor T

	// Calculate eigenvectors (Vi = columns of Q matrix)
	MechSys::tensor_to_matrix(T,M);
	if (LinAlg::JacobiRot(M,Q)==-1) return false;

	// Calculate eigenprojectors (Pi = Vi dyad Vi)
	for (int k=0; k<3; ++k)
	{
		// Eigenvalues
		ev[k] = M(k,k);
		// Eigenprojectors
		for (int i=0; i<3; ++i) ep[k][i] = Q(i,k) * Q(i,k);
		ep[k][3] = Q(0,k) * Q(1,k) * sq2;
		ep[k][4] = Q(1,k) * Q(2,k) * sq2;
		ep[k][5] = Q(0,k) * Q(2,k) * sq2;
	}
	return true;
}

// Square root of a tensor T:  sqrtT = T^(0.5)
bool MechSys::tensor_sqrt(const MechSys::Tensr2 & T,
		                        MechSys::Tensr2 & sqrtT)
{
	double          * ev = new double          [3]; // Eigenvalues
	MechSys::Tensr2 * ep = new MechSys::Tensr2 [3]; // Eigenprojectors

	// Calculate eigenvalues and eigenprojectors
	if (!MechSys::tensor_eigenvp(T,ev,ep)) return false;
	
	// Calculate sqrtT
	sqrtT.Clear();
	for (int i=0; i<3; ++i)
		sqrtT = sqrtT + ep[i] * sqrt(ev[i]);

	// Clean up
	delete [] ev;
	delete [] ep;
	return true;
}
