/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

#ifndef MECHSYS_MECHSYS_H
#define MECHSYS_MECHSYS_H

// external
#include "tensors/tensor1.h"
#include "tensors/tensor2.h"
#include "tensors/tensoperators.h"
#include "../linalg/matrix.h"
#include "../linalg/jacobirot.h"

// OBS.: Given the cartesian coordinates (T00, T11, T22, T01, T12, T02)
//       of a second order tensor, the Mandel representation is according
//       to:
//       Tensr2 = {T00 T11 T22 T01*sqrt(2) T12*sqrt(2) T02*sqrt(2)} ^(T)

namespace MechSys
{
	// Constants
	enum ModelID        {nlelastic=10 , subtij=20  , subcam=30 , wrongmodel=0};
	enum SchemeID       {FE=100       , ME=200     , wrongscheme=0};
	enum SubCamSurfType {Ori=1000     , Mod=2000};

	// Tensors
	typedef Tensor1<double,6> Tensr2; // 2nd tensor as V_6^1 (Mandel)
	typedef Tensor2<double,6> Tensr4; // 4th tensor as V_6^2 (Mandel)

	// Helpers
	MechSys::ModelID  GetModelID (const char * model);
	MechSys::SchemeID GetSchemeID(const char * scheme);
	
	// Functions
	void tensor_to_matrix(const Tensr2         & T, LinAlg::Matrix & M);
	void matrix_to_tensor(const LinAlg::Matrix & M, Tensr2         & T);

	// Tensor multiplication:  ST = S*T
	// -*- IMPORTANT -*-
	//       The function tensor_mult is valid only if the result R=S*T is 
	//       symmetric. This is NOT always true, even if S and T are both
	//       symmetric.
	void tensor_mult(const Tensr2 & S, const Tensr2 & T, Tensr2 & ST);
	
	// Inverse of a tensor T:  invT = T^(-1)
	bool tensor_inv(const Tensr2 & T, Tensr2 & invT);

	// Eigenvalues (ev) and Eigenprojectors (ep) of a tensor T
	bool tensor_eigenvp(const Tensr2 & T, double ev[3], Tensr2 ep[3]);

	// Square root of a tensor T:  sqrtT = T^(0.5)
	bool tensor_sqrt(const Tensr2 & T, Tensr2 & sqrtT);

}; // namespace MechSys

#endif
