/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

#include <iostream>
#include <cmath>
#include "../base/mechsys.h"
#include "../models/intvars.h"
#include "../models/nlelastic.h"
#include "../models/genericep.h"
#include "meuler.h"

namespace MechSys
{

bool SchModifiedEuler<NlElastic>::StressUpdate(NlElastic * model, const Tensr2 & Deps, Tensr2 & Dsig)
{
	std::cout << "NlElastic\n";
	std::cout << "ID = " << GetID() << std::endl;
	std::cout << "_sig = \n" << model->_sig << std::endl;
	_num_calls++;
	return true;
}

bool SchModifiedEuler<GenericEP>::StressUpdate(GenericEP * model, const Tensr2 & Deps, Tensr2 & Dsig)
{

	Tensr2  sig_ini(model->_sig); 
	IntVars   z_ini(model->_z.nIVLS(), model->_z.nIVns());
	//z_ini = model->_z;

	       _nss = 0;
	double    T = 0.0;
	double   dT = _dTini;

	Tensr2 de;  // strain increment with variable size
	Tensr2 ds1; // first stress increment
	Tensr2 ds2; // second stress increment
	Tensr2 sk;  // stress at the start of the iteration
	Tensr2 sc;  // correct stress
	Tensr2 se;  // erroneous stress
	
	int nIVLS = model->_z.nIVLS();
	int nIVns = model->_z.nIVns();

	IntVars dz1(nIVLS, nIVns); // first intvar increment
	IntVars dz2(nIVLS, nIVns); // second intvar increment
	IntVars zk (nIVLS, nIVns); // intvar at the start of the iteration
	IntVars zc (nIVLS, nIVns); // correct intvars
	IntVars ze (nIVLS, nIVns); // erroneous intvars

	double  err; // error
	double  m;   // multiplier for the next dT (new step-size)

	while (T<1.0 & _nss<=_max_it)
	{
		de = Deps * dT;
		sk = model->_sig;
		zk = model->_z;

		model->_dstress_given_dstrain(de,ds1,dz1);

		model->_sig = sk + ds1;
		  model->_z = zk + dz1;
		if (!model->_update_state()) return false;

		model->_dstress_given_dstrain(de,ds2,dz2);

		sc = sk + ds1 * 0.5 + ds2 * 0.5;
		zc = zk + dz1 * 0.5 + dz2 * 0.5;
		se = sc - model->_sig;
		ze = zc - model->_z;

		err = CalcNormS(se,ze) / CalcNormS(sc,zc);
		  m = 0.9*sqrt(_STOL/err);

		if (err<=_STOL) // step accepted
		{
			             T = T + dT;
			   model->_sig = sc;
			     model->_z = zc;
			if (m>_mMax) m = _mMax;
		}
		else           // step NOT accepted
		{
			   model->_sig = sk;
			     model->_z = zk;
			if (m<_mMin) m = _mMin;
		}

		if (!model->_update_state()) return false;

		dT = m * dT;
		if (dT>(1.0-T)) dT=1.0-T;

		_nss++;
	}

	// Statistics
	_num_calls++;
	_tot_nss = _tot_nss + _nss;
	if (_nss<_min_nss) _min_nss = _nss;
	if (_nss>_max_nss) _max_nss = _nss;
	
	if (_nss>_max_it)
	{
		std::cout << "~~~~~~~~~~~~~~ ERROR: ME didn't converge!! ~~~~~~~~~~~~~~\n";
		return false;
	}

	Dsig = model->_sig - sig_ini;
	//std::cout << "Dz = " << (model->_z - z_ini) << std::endl;
	
	return true;

}

}; // namespace MechSys
