/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

#include <iostream>
#include <cmath>
#include "jacobirot.h"
#include "matrix.h"

int LinAlg::JacobiRot(LinAlg::Matrix & A, LinAlg::Matrix & Q)
{
	const double   maxIt=30;       // Número máximo de iterações
	const double   errTol=1.0e-15; // Tolerância
	int            i;              // Contador (Linha)
	int            j;              // Contador (Coluna)
	int            r;              // Contador (O outro indice)
	double         rz;             // Razao (Arr-Ass)/2Ars
	double         c;              // Cosseno para a rot. sucessiva
	double         s;              // Seno
	double         cc;             // Cosseno ao quadrado
	double         ss;             // Seno ao quadrado
	double         t;              // Tangente do angulo para a rot.
	double         Temp;           // Variável auxiliar
	LinAlg::Matrix M(3,3);         // Matriz Auxiliar
	// Zerar Q
	Q(0,0)=1.0; Q(0,1)=0.0; Q(0,2)=0.0;
	Q(1,0)=0.0; Q(1,1)=1.0; Q(1,2)=0.0;
	Q(2,0)=0.0; Q(2,1)=0.0; Q(2,2)=1.0;
	// Rotacionar A
	double soma = fabs(A(0,1))+fabs(A(0,2))+fabs(A(1,2));
	int       k = 1;
	while (soma>errTol & k<=maxIt)
	{
		r = 2;              // Outro indice (em ordem decrescente)
		for (i=0; i<2; ++i) // DO i=1,2 ! Pimeiro indice
		{
			for (j=i+1; j<3; ++j) // DO j=i+1,3 ! Segundo indice
			{
				// Zerar termo i,j de A
				rz = (A(i,i)-A(j,j))/(2.0*A(i,j));
				if (fabs(rz)<0.0) t = 1.0;
				else t = 1.0/(rz+LinAlg::Sign(sqrt(rz*rz+1.0),rz));
				c  = 1.0/sqrt(1.0+t*t);
				s  = c*t;
				cc = c*c;
				ss = s*s;
				// Rotacionar A (zerar ij)
				Temp   = cc*A(i,i) + 2.0*c*s*A(i,j) + ss*A(j,j);
				A(j,j) = ss*A(i,i) - 2.0*c*s*A(i,j) + cc*A(j,j);
				A(i,i) = Temp;
				A(i,j) = 0.0;
				Temp   = c*A(i,r) + s*A(j,r);
				A(j,r) = c*A(j,r) - s*A(i,r);
				A(i,r) = Temp;
				A(j,i) = 0.0;
				A(r,i) = A(i,r);
				A(r,j) = A(j,r);
				//Atualizar Q
				M(i,i) =  Q(i,i)*c + Q(i,j)*s;
				M(j,i) =  Q(j,i)*c + Q(j,j)*s;
				M(r,i) =  Q(r,i)*c + Q(r,j)*s;
				Q(i,j) = -Q(i,i)*s + Q(i,j)*c;
				Q(j,j) = -Q(j,i)*s + Q(j,j)*c;
				Q(r,j) = -Q(r,i)*s + Q(r,j)*c;
				Q(i,i) =  M(i,i);
				Q(j,i) =  M(j,i);
				Q(r,i) =  M(r,i);
				// Próximo índice
				r--;
			}
		}
		soma = fabs(A(0,1))+fabs(A(0,2))+fabs(A(1,2));
		k++;
	}
	if (k>maxIt)
	{
		std::cout << "~~~~~~~~~~~~~~ ERROR: JacobiRot didn't converge!! ~~~~~~~~~~~~~~\n";
		return -1;
	}
	return k;
} // LinAlg::JacobiRot
