/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

// Numerical evaluation of eigenvalues and eigenvectors for 3x3 symmetric
// matrices using the Jacobi-rotation Method.
//
// 2005-05-10 (C): Dorival M. Pedroso (C++)
// 2004-08-15 (C): Dorival M. Pedroso (Fortran 95)
// 2002-04-15 (C): Dorival M. Pedroso (Fortran 77)
//
// INPUT
//        A: Matrix which eigenvalues and eigenvalues are to be found
// OUTPUT
//        A: Will be diagonal with eigenvalues along the diagonal
//        Q: Matriz which columns has the components of eigenvectors
// RETURN
//         k: if no problems, the number of iterations
//        -1: otherwise (thus, had problems!)
//
// Obs.: The matrix A must be symmetric

#ifndef LINALG_JACOBIROT_H
#define LINALG_JACOBIROT_H

#include "matrix.h"

namespace LinAlg
{
	
int JacobiRot(Matrix & A, Matrix & Q);

inline double Sign(const double x, const double y)
{
	/*
	The function SIGN in FORTRAN is called the sign transfer function. It
	is a function of two variables, and its definition involves two cases:

		CASE 1:	If y ≥ 0 then
				sign(x,y) = abs(x),
		CASE 2:	If y < 0 then
				sign(x,y) = - abs(x).

	The practical effect is that sign(x,y) has the same absolute value as
	x, but it has the same sign as y; thus the sign of y is transferred to
	x. (The case y = 0 is a little special - it gives sign(x,y) always a
	plus sign.)

	Examples:
		sign(2,3)=2, sign(2,-3)=-2, sign(-2,3)=2, sign(-2,-3)=-2.

	If we substitute x = 1 in the sign transfer function, we get the sign
	of y; that is,

		CASE 1:	If y ≥ 0 then
				sign(1,y) = 1,
		CASE 2:	If y < 0 then
				sign(1,y) = - 1.

	Thus, sign(1,y) in Fortran is essentially the same as the function
	SGN(y) in Basic (except when y = 0, when the Fortran value is +1 but
	the Basic value is 0).
	*/
	if (y >= 0.0) return fabs(x);
	else          return -fabs(x);
}

}; // namespace LinAlg

#endif
