/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

#ifndef MECHSYS_GENERICEP_H
#define MECHSYS_GENERICEP_H

#include <sstream>
#include "../base/mechsys.h"
#include "../parameters/parameters.h"
#include "../lawinteg/lawinteg.h"
#include "model.h"
#include "intvars.h"

namespace MechSys
{

template<typename ModelType> class SchForwardEuler;
template<typename ModelType> class SchModifiedEuler;

class GenericEP : public Model
{

	friend class SchForwardEuler <GenericEP>;
	friend class SchModifiedEuler<GenericEP>;

public:

	// Constructor
	GenericEP(int nIntVarLoadSurf, int nIntVarNormSurf)
		: _z(nIntVarLoadSurf, nIntVarNormSurf),
		  _is_simple_out(false),
		  _is_scheme_set(false)
	{
		 _ya = new double [nIntVarLoadSurf];
		_HHa = new double [nIntVarLoadSurf];
		_HHb = new double [nIntVarNormSurf];
	}

	// Destructor
	virtual ~GenericEP()
	{
		delete [] _ya;
		delete [] _HHa;
		delete [] _HHb;
	};

	// Inherited methods from Model
	bool     Set          (Parameters const * Prms, const Tensr2 & Sig);
	bool     Reset        (const Tensr2 & Sig);
	bool     Actualize    (const Tensr2 & Dsig, Tensr2 & Deps); // return Deps
	bool     StressUpdate (const Tensr2 & Deps, Tensr2 & Dsig); // return Dsig
	Tensr4 * TgCompliance (); // C
	Tensr4 * TgStiffness  (); // D
	void     Out          (std::ostringstream & oss) const;

	// Get
	int  GetnIVLS()              const { return _z.nIVLS(); }
	int  GetnIVns()              const { return _z.nIVns(); }
	void GetIntVars(IntVars & z) const { z = _z; }

	// Set
	void SetScheme(LawIntegScheme<GenericEP> * Scheme)
	{ _scheme = Scheme; _is_scheme_set = true; }
	void SetSimpleOut(bool IsSimpleOut)
	{ _is_simple_out = IsSimpleOut; }
		
protected:

	// Always up-to-date (fundamental)
	Tensr2   _sig; // Stress tensor
	IntVars  _z;   // Internal variables
	// Always up-to-date
	Tensr2   _r;   // Flow direction
	Tensr2   _v;   // v = df_dSig
	double * _ya;  // ya = df_dzalpha       (size=nIVLS)
	double * _HHa; // HHa = Halpha * halpha (size=nIVLS)
	double * _HHb; // HHb = Hbeta * hbeta   (size=nIVns)
	double   _cp;  // cp = - sum (yalpha * HHalpha)
	Tensr2   _b;   // b = (v:De)/(v:De:r+cp)  (dL = b:dEps)
	// Output
	bool     _is_simple_out;
	
	// Virtuals
	virtual bool _model_specific_set()        = 0;
	virtual bool _force_zalpha(Tensr2 & Sig)  = 0;
	virtual bool _start_intvars(Tensr2 & Sig) = 0;
	virtual bool _model_specific_update()     = 0;
	virtual bool _calculate_gradients()       = 0;
	virtual bool _calculate_hardening()       = 0;
	virtual void _model_specific_out(std::ostringstream & oss) const = 0;

private:

	// Constants
	Tensr4  _D;
	Tensr4  _C;
	double  _ce;
	// Always up-to-date
	double  _E;   // Young modulus
	Tensr4  _De;  // Elastic Stiffness
	// To be calculated
	Tensr4  _Dep;
	Tensr4  _Cep;
	// Scheme
	LawIntegScheme<GenericEP> * _scheme;
	bool                        _is_scheme_set;

	// Private functions
	bool _update_state(); // for previously set _sig and _z
	void _dstress_given_dstrain(const Tensr2 & Deps, Tensr2 & Dsig, IntVars & Dz);

}; // class GenericEP

}; // namespace MechSys

#endif
