/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

#ifndef INTVARS_H
#define INTVARS_H

#include <iostream>
#include <cmath>
#include <cassert>

namespace MechSys
{

class IntVars
{
public:
	IntVars(int nIntVarLoadSurf, int nIntVarNormalSurf)
		 : _nIVLS(nIntVarLoadSurf), _nIVns(nIntVarNormalSurf)
	{
		  _size = _nIVLS + _nIVns;
		_values = new double [_size];
		for (int i=0; i<_size; ++i) _values[i] = 0.0;
	}
	~IntVars() { delete [] _values; }

	// Get
	int    nIVLS()      const { return _nIVLS; }
	int    nIVns()      const { return _nIVns; }
	int    Size()       const { return _size;  }
	double EuclidNorm() const
	{
		double sum = 0.0;
		for (int i=0; i<_size; ++i)
			sum = sum + _values[i] * _values[i];
		return sqrt(sum);
	}

	// Set
	void Set(double * Va, double * Vb, double k)
	{
		for (int i=0; i<_nIVLS; ++i) _values[i]        = Va[i] * k;
		for (int i=0; i<_nIVns; ++i) _values[_nIVLS+i] = Vb[i] * k;
	}
	void Update(double * HHa, double * HHb, double dL)
	{
		for (int i=0; i<_nIVLS; ++i) _values[i]        = _values[i]        + HHa[i] * dL;
		for (int i=0; i<_nIVns; ++i) _values[_nIVLS+i] = _values[_nIVLS+i] + HHb[i] * dL;
	}
	
	// Operator Overloading
	void operator= (const IntVars & Rgt)
	{
		assert (&Rgt!=this);
		for (int i=0; i<_size; ++i) 
			_values[i] = Rgt._values[i];
	}
	double & operator[] (int i)
	{
		assert(i>=0 && i<_size);
		return _values[i];
	}
	const double & operator[] (int i) const
	{
		assert(i>=0 && i<_size);
		return _values[i];
	}

	// Binary operators
	IntVars operator+ (const IntVars & Rgt) const
	{
		IntVars res(_nIVLS,_nIVns);
		for (int i=0; i<_size; ++i) 
			res[i] = _values[i] + Rgt._values[i];
		return res;
	}
	IntVars operator- (const IntVars & Rgt) const
	{
		IntVars res(_nIVLS,_nIVns);
		for (int i=0; i<_size; ++i) 
			res[i] = _values[i] - Rgt._values[i];
		return res;
	}
	IntVars operator* (const double & Rgt) const
	{
		IntVars res(_nIVLS,_nIVns);
		for (int i=0; i<_size; ++i) 
			res[i] = _values[i] * Rgt;
		return res;
	}
	IntVars operator/ (const double & Rgt) const
	{
		IntVars res(_nIVLS,_nIVns);
		for (int i=0; i<_size; ++i) 
			res[i] = _values[i] / Rgt;
		return res;
	}

private:
	int      _nIVLS;
	int      _nIVns;
	int      _size;   // nIVLS+nIVns
	double * _values;
	
}; // class IntVars

std::ostream & operator<< (std::ostream & os, const IntVars & iv);

}; // namespace MechSys

#endif
