/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

#ifndef TIJTENSOR_H
#define TIJTENSOR_H

#include <string>
#include <cassert>
#include "../base/mechsys.h"

namespace MechSys
{

class TijTensor
{
public:
	TijTensor(double TsMin);
	bool Set(const Tensr2 & Sig);
	bool CalcDerivs();

	double tn() const { assert(_is_stress_ok); return _tn; }
	double ts() const { assert(_is_stress_ok); return _ts; }
	double rz() const { assert(_is_stress_ok); return _rz; }
	double I1() const { assert(_is_stress_ok); return _I1; }
	double I2() const { assert(_is_stress_ok); return _I2; }
	double I3() const { assert(_is_stress_ok); return _I3; }
	
	const Tensr2 * tau()     const { assert(_is_stress_ok); return (&_tau); }
	const Tensr2 * inv_tau() const { assert(_is_stress_ok); return (&_inv_tau); }
	const Tensr2 * aij()     const { assert(_is_stress_ok); return (&_aij); }
	const Tensr2 * tij()     const { assert(_is_stress_ok); return (&_tij); }
	const Tensr2 * dtn_ds()  const { assert(_is_derivs_ok); return (&_dtn_ds); }
	const Tensr2 * dts_ds()  const { assert(_is_derivs_ok); return (&_dts_ds); }
	const Tensr2 * dts_dt()  const { assert(_is_derivs_ok); return (&_dts_dt); }

	std::string GetErrors() const { return _errors; }

private:
	double  _ts_min; // minimum value for ts

	// Stress
	Tensr2  _sig;
	double  _I1;
	double  _I2;
	double  _I3;
	double  _tn;
	double  _ts;
	double  _rz;          // _rz = sqrt(I3/I2)
	Tensr2  _tau;         // _tau = sqrt(Sig)
	Tensr2  _inv_tau;     // _inv_tau = _tau^(-1)
	Tensr2  _aij;         // _aij = _inv_tau * _rz
	Tensr2  _tij;         // _tij = _aij * _sig
	bool    _is_stress_ok;

	// Derivatives
	Tensr2  _dI0_ds;
	Tensr2  _dI1_ds;
	Tensr2  _dI2_ds;
	Tensr2  _dtn_ds;
	Tensr2  _dts_ds;
	Tensr2  _dts_dt;
	bool    _is_derivs_ok;

	// Errors
	std::string _errors;

}; // class TijTensor
	
}; // namespace MechSys

#endif
