/*************************************************************************************
 * MechSys - A C++ library to simulate (Continuum) Mechanical Systems                *
 * Copyright (C) 2005 Dorival de Moraes Pedroso <dorival_pedroso em yahoo.com>       *
 *                                                                                   *
 * This file is part of MechSys.                                                     *
 *                                                                                   *
 * MechSys is free software; you can redistribute it and/or modify it under the      *
 * terms of the GNU General Public License as published by the Free Software         *
 * Foundation; either version 2 of the License, or (at your option) any later        *
 * version.                                                                          *
 *                                                                                   *
 * MechSys is distributed in the hope that it will be useful, but WITHOUT ANY        *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A   *
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.          *
 *                                                                                   *
 * You should have received a copy of the GNU General Public License along with      *
 * MechSys; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, *
 * Fifth Floor, Boston, MA 02110-1301, USA                                           *
 *************************************************************************************/

#include <iostream>
#include <cassert>
#include <cmath>
#include "../base/mechsys.h"
#include "parameters.h"

namespace MechSys
{

Parameters::Parameters()
{
	_lam=0.0; _kap=0.0;   _nu=-1.0;
	_rcs=0.0; _enc=0.0;	 _pnc=0.0; _bet=0.0;
	 _c1=0.0;  _c2=0.0;   _d1=0.0;  _d2=0.0;
	
	_is_vri_calc = false; _vri=0.0;
	_is_M_calc   = false;   _M=0.0;
	_is_mus_calc = false; _xcs=0.0; _ycs=0.0; _mus=0.0;
}

void Parameters::Calculatevri(Tensr2 & Sig)
{
	assert(_lam>0.0);
	assert(_enc>0.0);
	assert(_pnc>0.0);
	_vri = _enc + _lam * log(3.0*_pnc/(Sig[0]+Sig[1]+Sig[2]));
	_is_vri_calc = true;
}

void Parameters::CalculateM()
{
	assert(_rcs>0.0);
	_M = (3.0*_rcs-3.0)/(_rcs+2.0);
	_is_M_calc = true;
}

void Parameters::CalculateMus()
{
	assert(_rcs>0.0);
	assert(_bet>0.0);
	_xcs = (sqrt(_rcs)-1.0/sqrt(_rcs))*sqrt(2.0)/3.0;
	_ycs = ((1.0-sqrt(_rcs))/(0.5+sqrt(_rcs)))/sqrt(2.0);
	_mus = pow(pow(_xcs,_bet)+_ycs*pow(_xcs,(_bet-1.0)),1.0/_bet);
	_is_mus_calc = true;
}

void Parameters::_calculate_phics()
{
	assert(_rcs>0.0);
	_phics = asin((_rcs-1.0)/(_rcs+1.0))*45.0/atan(1.0);
}

}; // namespace MechSys

std::ostream & operator<< (std::ostream & os, const MechSys::Parameters & mp)
{
	if (mp._lam>0.0)        os << "  lam = " << mp._lam   << std::endl;
	if (mp._kap>0.0)        os << "  kap = " << mp._kap   << std::endl;
	if (mp._nu>=0.0)        os << "   nu = " << mp._nu    << std::endl;
	if (mp._rcs>0.0)        os << "  rcs = " << mp._rcs   << std::endl;
	if (mp._enc>0.0)        os << "  enc = " << mp._enc   << std::endl;
	if (mp._pnc>0.0)        os << "  pnc = " << mp._pnc   << std::endl;
	if (mp._rcs>0.0)        os << "phics = " << mp._phics << std::endl;
	if (mp._is_M_calc)      os << "    M = " << mp._M     << std::endl;
	if (mp._is_vri_calc)    os << "  vri = " << mp._vri   << std::endl;
	if (mp._bet>0.0)        os << "  bet = " << mp._bet   << std::endl;
	if (mp._is_mus_calc)    os << "  xcs = " << mp._xcs   << std::endl;
	if (mp._is_mus_calc)    os << "  ycs = " << mp._ycs   << std::endl;
	if (mp._is_mus_calc)    os << "  mus = " << mp._mus   << std::endl;
	if (mp._c1>0.0)         os << "   c1 = " << mp._c1    << std::endl;
	if (mp._c2>0.0)         os << "   c2 = " << mp._c2    << std::endl;
	if (mp._d1>0.0)         os << "   d1 = " << mp._d1    << std::endl;
	if (mp._d2>0.0)         os << "   d2 = " << mp._d2    << std::endl;

	return os;
}
